# coding=utf8
from typing import Dict, Union

from pydantic import BaseModel

from ..Entities import XLLinkDE


class HttpFuncReq(BaseModel):
    func_key: str
    data_dict: Dict[str, XLLinkDE]
    env_dict: dict = {}
    pass


class HttpEventReq(BaseModel):
    event: str
    data_dict: Dict[str, XLLinkDE]
    want_keys: Union[list] = []
    env_dict: dict = {}
    pass


class HttpLearnReq(BaseModel):
    context_id: str
    event: str
    data_dict: Dict[str, XLLinkDE]
    env_dict: dict = {}
    pass


class XLResponse(object):
    def __init__(self, code, message):
        self.code = code
        self.message = message
        self.data_dict = {}
        self.data_dict_json = {}
        self.env_dict = {}

    def set_covert_data_dict(self, data_dict: dict):
        data_dict_json = {k: d.toLinkDE() for k, d in data_dict.items()}
        self.data_dict = data_dict
        self.data_dict_json = data_dict_json
        return self

    def to_app_back(self) -> dict:
        return {
            'code': self.code,
            'message': self.message,
            'env_dict': self.env_dict,
            'data_dict': self.data_dict_json
        }

    pass



# coding=utf8

from .Storage import XLCGlobalInfo, XLContext
from .XLDataStructure import XLDE, XLDataUtil, XLLinkDE
from .XLDataMark import XLDataMarkUtil, XLDM, XLLinkDM
from .RunEntiy import RunTraceE
from .XLExceptionEntities import XLBaseException, XLExceptionEnum




# coding=utf8
"""
logging模块是Python的内置模块，不需要安装。

步骤：
1，创建一个把日志信息保存到文件中的处理器FileHandler
2，把文件处理器添加到logger中
3，把格式器传入到文件处理器中
"""
# 导入logging模块
import logging.handlers
import os

import logstash

from ..Department import BaseDPart
from ..XLSYS.Base import SingletonArgs
from ..XLSYS.Entities import XLContext


class MyLogHandler(logging.Handler, object):
    def __init__(self, extend_info=None, department_name=None, env_flag=None, **kwargs):
        logging.Handler.__init__(self)
        self.base_info = {
            'department': department_name,
            'env_flag': env_flag
        }
        if extend_info is not None and isinstance(extend_info, dict):
            self.base_info.update(extend_info)
        pass

    def emit(self, record):
        # 注意不要大写
        record.__dict__.update(self.base_info)


def buildRequestLogger(log_dir, department_name, env_flag):
    requestLogger = logging.getLogger('request')
    requestLogger.setLevel(level=logging.INFO)
    # 添加TimedRotatingFileHandler
    # 定义一个1秒换一次log文件的handler
    # 保留5个旧log文件
    timefilehandler = logging.handlers.TimedRotatingFileHandler(
        '{}/request_{}.log'.format(log_dir, department_name),
        when='D',
        interval=1,
        encoding="utf-8",
        backupCount=5
    )

    # 设置后缀名称，跟strftime的格式一样
    timefilehandler.suffix = "%Y-%m-%d.log"
    # 创建日志格式器
    formatter = logging.Formatter(
        fmt="%(asctime)s| [%(mod)s] | [%(context_id)s| [%(global_info)s |\n%(message)s")
    timefilehandler.setFormatter(formatter)
    requestLogger.addHandler(timefilehandler)
    if env_flag == 'dev':
        chlr = logging.StreamHandler()  # 输出到控制台的handler
        chlr.setFormatter(formatter)
        requestLogger.addHandler(chlr)

    return requestLogger
    pass


@SingletonArgs
class XLLogger(logging.getLoggerClass()):
    def __init__(self, department: BaseDPart = None, base_path=None, logstash_host=None, env_flag='dev',
                 extend_info=None, **kwargs):
        logging.getLoggerClass().__init__(self, "logger", level=logging.INFO)
        department_name = department.name.lower().strip()
        my_log_handler = MyLogHandler(extend_info=extend_info, department_name=department_name, env_flag=env_flag,
                                      **kwargs)
        # 如果日志文件夹不存在，则创建
        if base_path is not None and isinstance(base_path, str):
            log_path = base_path if base_path.endswith("/") else base_path + "/"
            log_dir = log_path + "logs"
        else:
            log_dir = os.path.abspath('..') + os.sep + "logs"  # 日志存放文件夹名称
        if not os.path.isdir(log_dir):
            os.makedirs(log_dir)
        # 添加TimedRotatingFileHandler
        # 定义一个1秒换一次log文件的handler
        # 保留5个旧log文件
        timefilehandler = logging.handlers.TimedRotatingFileHandler(
            '{}/log_{}.log'.format(log_dir, department_name),
            when='D',
            interval=1,
            encoding="utf-8",
            backupCount=5
        )

        # 设置后缀名称，跟strftime的格式一样
        timefilehandler.suffix = "%Y-%m-%d.log"
        # 创建日志格式器
        formatter = logging.Formatter(
            fmt="%(asctime)s [%(filename)s %(lineno)d行]| [ %(levelname)s ] | [%(message)s]")
        timefilehandler.setFormatter(formatter)
        self.addHandler(timefilehandler)
        self.addHandler(my_log_handler)
        if logstash_host is not None:
            self.addHandler(logstash.LogstashHandler(logstash_host, 20044, version=1))

        if env_flag == 'dev':
            chlr = logging.StreamHandler()  # 输出到控制台的handler
            chlr.setFormatter(formatter)
            self.addHandler(chlr)

        self.requestLogger = buildRequestLogger(log_dir, department_name, env_flag)
        pass

    def infoContext(self, context: XLContext, msg, **kwargs):
        extra = {
            'context_id': context.context_id,
            'all_trace_id': context.global_info.all_trace_id
        }
        self.info(msg, extra=extra, **kwargs)
        pass

    def errorContext(self, context: XLContext, msg, **kwargs):
        extra = {
            'context_id': context.context_id,
            'all_trace_id': context.global_info.all_trace_id
        }
        self.error(msg, extra=extra, **kwargs)
        pass



# coding=utf8

__OBJECT_POOL = {}
__OBJECT_INFO_DICT = {}


def getObj(name_key, prn=True):
    _obj = __OBJECT_POOL.get(name_key)
    if prn and _obj is None:
        print('getObj 为空: ',name_key)
    return _obj
    pass


def getObjInfo(name_key) -> dict:
    return __OBJECT_INFO_DICT.get(name_key)
    pass


def updateObj(name_key, obj, info={}):
    __OBJECT_POOL[name_key] = obj
    __OBJECT_INFO_DICT[name_key] = info
    pass








# coding=utf8
import inspect
from importlib import import_module

from . import XLObjectUtil


class AISerializeObject(object):

    def __to_json__(self):
        pass

    @staticmethod
    def __from_json__(json_str: str):
        pass

    pass


class AIDataSerialize(AISerializeObject):
    def __to_json__(self):
        save_dict = {'__class_data': self.__dict__, '__class_flag': "AISerializeObject", '__class_module': self.__module__,
                     '__class_name': self.__class__.__name__}
        return save_dict
        pass

    @staticmethod
    def __from_json__(json_dict: dict):
        obj_class_info = getattr(import_module(name=json_dict['__class_module']), json_dict['__class_name'])
        obj = object.__new__(obj_class_info)
        obj.__dict__ = json_dict['__class_data']
        return obj
        pass

    pass


class AIFuncSerialize(AISerializeObject):
    def __to_json__(self):
        save_dict = {'__class_data': self.__dict__, '__class_flag': "AISerializeObject", '__class_module': self.__module__,
                     '__class_name': self.__class__.__name__}
        return save_dict
        pass

    @staticmethod
    def __from_json__(json_dict: dict):
        obj_class_info = getattr(import_module(name=json_dict['__class_module']), json_dict['__class_name'])
        obj = object.__new__(obj_class_info)
        obj.__dict__ = json_dict['__class_data']
        return obj

    pass


class ObjectMangeBase(object):

    def __init__(self, name_key, cover_obj=False, **kwargs):
        self.name_key = name_key
        old_obj = XLObjectUtil.getObj(name_key, prn=False)
        if old_obj is not None and not cover_obj:
            # old_obj_info = XLObjectUtil.getObjInfo(name_key)
            # print(json.dumps(old_obj_info))
            raise Exception("{} 的 obj已经存在了".format(name_key))
        # 获取当前调用栈的帧
        stack = inspect.stack()
        obj_info = {
            'stack_info_list': ["{}--{}".format(stack[i].filename, stack[i].lineno) for i in range(1, len(stack)) if
                                "XLAI" in stack[i].filename]
        }

        XLObjectUtil.updateObj(name_key, self, obj_info)
        pass

    pass



# coding=utf8

class BaseDPart(object):

    def __init__(self, name, host, port):
        self.name = name
        self.host = host
        self.port = port
        pass

    def getHttpUrl(self, path):
        if path.startswith('/'):
            path = path[1:]
        return "http://{}:{}/{}".format(self.host, self.port, path)

    pass


PP = BaseDPart('PP', 'pp.ai.home.tokgo.cn', 15001)
SF = BaseDPart('SF', 'sf.ai.home.tokgo.cn', 15002)
AL = BaseDPart('AL', 'al.ai.home.tokgo.cn', 15003)
MB = BaseDPart('MB', 'mb.ai.home.tokgo.cn', 15004)
EXT = BaseDPart('EXT', 'ext.ai.home.tokgo.cn', 15005)
EV = BaseDPart('EV', 'ev.ai.home.tokgo.cn', 15006)
MP = BaseDPart('MP', 'mp.ai.home.tokgo.cn', 15007)



# coding=utf8
import traceback
from enum import Enum


class XLExceptionEnum(Enum):
    http_debug_timeout = "http_debug_timeout"

    pass


class XLBaseException(Exception):
    def __init__(self, event: XLExceptionEnum, des="", **kwargs):
        self.event = event
        self.stack = traceback.extract_stack()
        self.kwargs = kwargs
        self.des = des
        self.info_dict = {}
        self.data_dict = {}
        pass

    def printStack(self):
        traceback.print_list(self.stack)
        pass

    def updateEXDataList(self, data_list: list):
        for d in data_list:
            self.data_dict[d.key] = d
        return self
        pass

    def addInfoList(self, info_dict: dict):
        self.info_dict.update(info_dict)
        return self
        pass


pass



# coding=utf8
import json
from importlib import import_module
from io import StringIO
import pandas
import torch
import numpy
from datetime import datetime
from .XLRedisUtil import XLRedisUtil
from expiringdict import ExpiringDict
from ..XLSYS.Base import AISerializeObject
from ..XLSYS.Entities import XLContext


class ContextSaveEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            save_dict = {'__class_flag': "datetime", '__time': obj.strftime("%Y-%m-%d %H:%M:%S")}
            return save_dict
        elif isinstance(obj, numpy.ndarray):
            save_dict = {'__class_flag': "numpy", '__data': json.dumps(obj.tolist(), ensure_ascii=False)}
            return save_dict
        elif isinstance(obj, pandas.DataFrame):
            save_dict = {'__class_flag': "DataFrame", '__data': obj.to_json()}
            return save_dict
        elif isinstance(obj, torch.Tensor):
            save_dict = {'__class_flag': "torch", '__data': json.dumps(obj.detach().cpu().tolist(), ensure_ascii=False)}
            return save_dict
        if isinstance(obj, bytes):
            return str(obj, encoding='utf-8')
        elif isinstance(obj, AISerializeObject):
            save_dict = obj.__to_json__()
            return save_dict
        else:
            raise Exception("ContextSaveEncoder Unexpected object type " + str(type(obj)))


class ContextLoadDecoder(json.JSONDecoder):

    def __init__(self, *args, **kwargs):
        super().__init__(
            object_hook=self.object_hook,
            *args, **kwargs)

    def object_hook(self, o):
        """
        对字典解码，如果同时设置'object_pairs_hook'，'object_hook'将不生效
        :param o: 原始数据字典
        :return: 实际需要的数据
        """
        if o.get('__class_flag') is not None and o.get('__class_flag') == 'numpy':
            return numpy.asarray(json.loads(o['__data']))
        if o.get('__class_flag') is not None and o.get('__class_flag') == 'torch':
            return torch.asarray(numpy.asarray(json.loads(o['__data'])))
        if o.get('__class_flag') is not None and o.get('__class_flag') == 'DataFrame':
            return pandas.read_json(StringIO(o['__data']))
        if o.get('__class_flag') is not None and o.get('__class_flag') == 'AISerializeObject':
            return getattr(import_module(name=o['__class_module']), o['__class_name']).__from_json__(o)
        if o.get('__class_flag') is not None and o.get('__class_flag') == 'datetime':
            return datetime.strptime(o['__time'], "%Y-%m-%d %H:%M:%S")
        return o


class XLStorageUtil(object):
    def __init__(self, max_len=1000, pool_ttl_seconds=3600):
        self.pool_ttl_seconds = pool_ttl_seconds
        self.redis_util = XLRedisUtil()
        # 创建一个带有过期时间的字典，过期时间为5秒
        self.context_pool = ExpiringDict(max_len=max_len, max_age_seconds=pool_ttl_seconds)
        pass

    def getFlow(self, context_id) -> XLContext:
        local_context = self.context_pool.get(context_id)
        if local_context is not None:
            return local_context
        redis_data_str = self.redis_util.get("context:" + context_id)
        if redis_data_str is None:
            return None
        redis_context = json.loads(redis_data_str, cls=ContextLoadDecoder)
        if redis_context is None:
            return None
        self.context_pool[context_id] = redis_context
        return redis_context
        pass

    def saveContextInfo(self, context: XLContext):
        save_info = json.dumps(context, cls=ContextSaveEncoder, ensure_ascii=False)
        self.redis_util.set("context:" + context.context_id, save_info, ex=self.pool_ttl_seconds)
        pass


pass



# coding=utf8
from typing import List

from ..Base import AISerializeObject


class RunTraceE(AISerializeObject):
    def __init__(self, name_key):
        self.name_key = name_key
        self.input_data_torch = None
        self.out_data = None
        self.next_trace_list: List[RunTraceE] = []
        pass

    pass




# coding=utf8
import copy
import json
from datetime import datetime
from io import StringIO
from typing import Optional, List

import numpy
import pandas
from pydantic import BaseModel

from . import XLContext
from ..Base import AIDataSerialize


class XLLinkDE(BaseModel):
    key: str
    data_type: str
    data: object
    encode_type: Optional[str] = ''
    encode_data_off: Optional[int] = 0
    trace_info: dict


class XLDE(AIDataSerialize):
    def __init__(self, key, data_type, data, encode_type='',
                 encode_data_off=0, trace_info=None):
        # 类型
        self.data_type = data_type
        # key
        self.key = key
        # 数据
        self.data = data
        self.trace_info: dict = {} if trace_info is None else trace_info
        # 是否运行编码
        self.encode_type = encode_type
        self.encode_data_off = encode_data_off
        self.encode_list = []

        self.covertData()
        pass

    def setTrace_info(self, link_id, up_trace_info=None):
        self.trace_info['newest_link_id'] = link_id
        if "link_id_list" in self.trace_info:
            self.trace_info.get("link_id_list").append(link_id)
        else:
            self.trace_info['link_id_list'] = [link_id]
        if up_trace_info is not None:
            self.trace_info['up_trace_info'] = copy.deepcopy(up_trace_info)
        return self
        pass

    def __str__(self):
        if self.data_type == 'datetime' and isinstance(self.data, datetime):
            str_data = self.data.strftime("%Y-%m-%d %H:%M:%S")
        elif self.data_type == 'ndarray' and isinstance(self.data, numpy.ndarray):
            # 将numpy数组转换为列表
            data_list = self.data.tolist()
            # 将数据转换为JSON格式的字符串
            str_data = json.dumps(data_list, ensure_ascii=False)
        elif self.data_type == 'DataFrame':
            str_data = 'DataFrame'
        else:
            str_data = self.data
        return '<{}>{}'.format(self.key, str_data)
        pass

    def covertData(self):
        if self.data_type == 'datetime' and isinstance(self.data, str):
            self.data = datetime.strptime(self.data, "%Y-%m-%d %H:%M:%S")
        if self.data_type == 'ndarray' and isinstance(self.data, str):
            self.data = numpy.asarray(json.loads(self.data))
        if self.data_type == 'DataFrame' and isinstance(self.data, str):
            from ...Util import XLRedisUtil
            data_str = XLRedisUtil().get_data_client().get(self.data)
            if data_str is not None:
                self.data = pandas.read_json(StringIO(data_str))
            else:
                self.data_type = 'expire'
            pass
        pass

    def toLinkDE(self):
        if self.data_type == 'datetime' and isinstance(self.data, datetime):
            str_data = self.data.strftime("%Y-%m-%d %H:%M:%S")
        elif self.data_type == 'ndarray' and isinstance(self.data, numpy.ndarray):
            # 将numpy数组转换为列表
            data_list = self.data.tolist()
            # 将数据转换为JSON格式的字符串
            str_data = json.dumps(data_list, ensure_ascii=False)
        elif self.data_type == 'DataFrame':
            from ...Util import XLRedisUtil
            str_data = datetime.now().strftime("ext-data-DataFrame-%Y%m%d_%H%M%S_%f")
            # 默认10分钟 为了debug 先弄一天
            XLRedisUtil().get_data_client().set(str_data, self.data.to_json(), ex=86400)
        else:
            str_data = self.data

        return {
            'key': self.key,
            'data_type': self.data_type,
            'encode_type': self.encode_type,
            'encode_data_off': self.encode_data_off,
            'data': str_data,
            'trace_info': copy.deepcopy(self.trace_info)
        }
        pass


class XLDataUtil(object):

    @staticmethod
    def INPUT_Build(data_info: XLLinkDE) -> XLDE:
        entity: XLDE = XLDE(key=data_info.key, data_type=data_info.data_type, data=data_info.data,
                            encode_type=data_info.encode_type,
                            encode_data_off=data_info.encode_data_off,
                            trace_info=copy.deepcopy(data_info.trace_info))
        return entity
        pass

    @staticmethod
    def SELF_Build(key, data, context: XLContext, data_type=None, up_trace_info=None) -> XLDE:
        if data_type is None:
            data_type = type(data).__name__
        entity: XLDE = XLDE(key=key, data_type=data_type, data=data)
        entity.setTrace_info(context.context_id, up_trace_info=up_trace_info)
        return entity
        pass

    @staticmethod
    def STEP_Build(data_info: dict, context: XLContext = None, up_trace_info=None) -> XLDE:
        data_type = data_info.get('data_type')
        data = data_info.get('data')
        encode_type = '' if data_info.get('encode_type') is None else data_info.get('encode_type')
        encode_data_off = 0 if data_info.get('encode_data_off') is None else data_info.get('encode_data_off')

        entity: XLDE = XLDE(key=data_info.get('key'), data_type=data_type, data=data, encode_type=encode_type,
                            encode_data_off=encode_data_off, trace_info=copy.deepcopy(data_info.get('trace_info')))
        if context is not None:
            entity.setTrace_info(context.context_id, up_trace_info=up_trace_info)
        return entity
        pass

    @staticmethod
    def FUNC_Build(key, data, link_id, up_trace_info=None) -> XLDE:
        data_type = type(data).__name__
        entity: XLDE = XLDE(key=key, data_type=data_type, data=data)
        entity.setTrace_info(link_id, up_trace_info=up_trace_info)
        return entity
        pass

    @staticmethod
    def LIST_DE_Build(key, data_de_list: list, context: XLContext, up_trace_info=None) -> XLDE:
        run_data_de_list = []
        for data_de in data_de_list:
            if isinstance(data_de, XLDE):
                run_data_de_list.append(data_de.toLinkDE())
            elif isinstance(data_de, XLLinkDE):
                run_data_de_list.append(data_de.__dict__)
            else:
                run_data_de_list.append(data_de)

        data = [de.get('data') for de in run_data_de_list]
        entity: XLDE = XLDE(key=key, data_type="xl_list", data=data)
        entity.setTrace_info(context.context_id, up_trace_info=up_trace_info)
        entity.trace_info['data_remark_info'] = [{"trace_info": copy.deepcopy(de.get('trace_info')), "data_type": de.get('data_type')} for de in data_de_list]
        return entity
        pass


pass



# coding=utf8
import json
from datetime import datetime
from importlib import import_module

from ..XLSYS.Base import AISerializeObject


class CommonEncoderALL(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S")
        if isinstance(obj, bytes):
            return str(obj, encoding='utf-8')
        if isinstance(obj, int):
            return int(obj)
        elif isinstance(obj, float):
            return float(obj)
        elif isinstance(obj, AISerializeObject):
            return obj.__dict__
        else:
            ""


class CommonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S")
        if isinstance(obj, bytes):
            return str(obj, encoding='utf-8')
        if isinstance(obj, int):
            return int(obj)
        elif isinstance(obj, float):
            return float(obj)
        elif isinstance(obj, object):
            return obj.__str__()
        else:
            return super(CommonEncoder, self).default(obj)


class SaveEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            save_dict = {'__class_flag': "datetime",  '__time': obj.strftime("%Y-%m-%d %H:%M:%S")}
            return save_dict
        if isinstance(obj, bytes):
            return str(obj, encoding='utf-8')
        elif isinstance(obj, object):
            save_dict = {'__class_data': obj.__dict__, '__class_flag': "class", '__class_module': obj.__module__,
                         '__class_name': obj.__class__.__name__}
            return save_dict
        else:
            return super(SaveEncoder, self).default(obj)


class LoadDecoder(json.JSONDecoder):

    def __init__(self, *args, **kwargs):
        super().__init__(
            object_hook=self.object_hook,
            # parse_float=self.parse_float,
            # parse_int=self.parse_int,
            # parse_constant=self.parse_constant,
            # object_pairs_hook=self.object_pairs_hook,
            *args, **kwargs)

    def object_hook(self, o):
        """
        对字典解码，如果同时设置'object_pairs_hook'，'object_hook'将不生效
        :param o: 原始数据字典
        :return: 实际需要的数据
        """
        if o.get('__class_flag') is not None and o.get('__class_flag') == 'class':
            obj = getattr(import_module(name=o['__class_module']), o['__class_name'])()
            obj.__dict__ = o['__class_data']
            return obj
        if o.get('__class_flag') is not None and o.get('__class_flag') == 'datetime':
            return datetime.strptime(o['__time'], "%Y-%m-%d %H:%M:%S")
        return o

    # def parse_float(self, o):
    #     """
    #     对浮点型解码
    #     :param o: 原始浮点型
    #     :return: 实际需要的数据
    #     """
    #     return o
    #
    # def parse_int(self, o):
    #     """
    #     对整型解码
    #     :param o: 原始整型
    #     :return: 实际需要的数据
    #     """
    #     return o
    #
    # def parse_constant(self, o):
    #     """
    #     对于'-Infinity', 'Infinity', 'NaN'之类的特殊值进行解码
    #     :param o:
    #     :return:
    #     """
    #     return o
    #
    # def object_pairs_hook(self, o):
    #     """
    #     对有序列表解码，如果同时设置'object_hook'，'object_hook'将不生效
    #     :param o: 原始数据数据
    #     :return: 实际需要的数据
    #     """
    #     return o



# coding=utf8
import inspect
import json
import time
import traceback
from concurrent.futures import ThreadPoolExecutor

import requests
from fastapi.dependencies.utils import analyze_param
from redis import Redis

from ..Entities import XLContext
from ...Department import BaseDPart
from ...Util import XLRedisUtil, XLLogger

_register_link = []


def mqLink(path, **kwargs):
    def decorator(func):
        sig = inspect.signature(func)
        param_field_list = []
        for param_name, param in sig.parameters.items():
            _param_info = analyze_param(
                param_name=param_name,
                annotation=param.annotation,
                value=param.default,
                is_path_param=False,
            )
            param_field_list.append(_param_info.field)
        _register_link.append([path, kwargs, func, param_field_list])

    return decorator


def _getSetSubIndex(redis_client: Redis, name, group, new_id=None):
    key = group + ":" + name + "readid"
    if new_id is not None:
        redis_client.set(key, new_id)
        return new_id
    return redis_client.get(key) or '0-0'


def _mqHandler(link_config: list, env_flag):
    logger = XLLogger()
    try:
        redis_util = XLRedisUtil()
        [mq_key, kwargs, func, param_list] = link_config
        if env_flag == 'dev':
            mq_key = 'dev_' + mq_key
        group = 'main'
        redis_client = redis_util.get_mq_client()
        try:
            redis_client.xinfo_consumers(name=mq_key, groupname=group)
        except Exception as e:
            redis_client.xadd(mq_key, {'sys': 'create'})
            consumer_id = _getSetSubIndex(redis_client, group=group, name=mq_key)
            redis_client.xgroup_create(name=mq_key, groupname=group, id=consumer_id)
        # logger.info('mq 启动成功 mq_key:{}'.format(mq_key))
        while True:
            content = {}
            try:
                # block 0 时阻塞等待, 其他数值表示读取超时时间
                [[name, [(nid, content)]]] = redis_client.xreadgroup(groupname=group, consumername="consumer",
                                                                     streams={mq_key: '>'}, count=1, block=0)
                _ = _getSetSubIndex(redis_client, group=group, name=mq_key, new_id=nid)
                if not isinstance(content, dict) or len(content) == 0 or content.get('sys') is not None:
                    continue
                if env_flag != 'dev' and redis_util.get('mq_dev_flag') is not None:
                    logger.info("mq base 数据转发 mq_key：{} ttl:{}".format(mq_key, redis_util.ttl('mq_dev_flag')))
                    _ = redis_client.xadd('dev_' + mq_key, content)
                    continue
                    pass
                content = {k: json.loads(v.encode('utf-8')) for k, v in content.items()}
                run_args = {}
                for param in param_list:
                    body_values, errors_ = param.validate(content[param.name], {}, loc=("body",))
                    if isinstance(errors_, list) or errors_:
                        logger.error('mq base 解析数据错误； key:{} , data:{}'.format(param.name, content[param.name]))
                        continue
                    run_args[param.name] = body_values
                    pass
                logger.info("mq base 收到消息; mq_key:{}| {} ".format(mq_key, json.dumps(content['req'], ensure_ascii=False)))
                func(**run_args)
                pass
            except Exception as ex:
                logger.error(" mq base  mq消息处理异常 content:{} e:{}".format(content, ex))
                traceback.print_exc()
                pass
    except Exception as e:
        logger.error("mq base  mq _mqRun 启动异常 link_config:{} e:{}".format(link_config, e))
        traceback.print_exc()
    pass


def _mqdev_handler(department: BaseDPart, env_flag):
    logger = XLLogger()
    if env_flag != 'dev':
        # logger.info('_mqdev_handler 环境不是dev 可以不用执行')
        return
    while True:
        try:
            response = requests.post('http://host.home.tokgo.cn:{}/mq/dev/link/register'.format(department.port))
            # 间隔50秒
            time.sleep(50)
            pass
        except Exception as e:
            logger.error('_mqdev_handler 处理任务错')
            traceback.print_exc()
            time.sleep(30)
            pass
    pass


def sendMqData(mq_key, req, context: XLContext = None):
    logger = XLLogger()
    try:
        t_data = json.dumps(req, ensure_ascii=False)
        if context is not None:
            extra = {"mod": "mq", "context_id": context.context_id,
                     "global_info": context.global_info.to_dict()}
        else:
            extra = {"mod": "mq", "context_id": "", "global_info": ""}
        # 打印请求日志
        logger.requestLogger.info("{} {}".format(mq_key, t_data), extra=extra)
        return XLRedisUtil().get_mq_client().xadd(mq_key, {"req": t_data})
    except Exception as e:
        logger.error("mq 发送消息失败; e:{}".format(e))
    pass


def mqSystemInit(department: BaseDPart, env_flag='dev'):
    if len(_register_link) == 0:
        return
    threadPool = ThreadPoolExecutor(max_workers=len(_register_link) + 1)
    for link_config in _register_link:
        threadPool.submit(_mqHandler, link_config, env_flag)

    # dev 环境相关的处理线程
    threadPool.submit(_mqdev_handler, department, env_flag)
    pass



# coding=utf8
import redis
import json

from ..Department import BaseDPart
from ..XLSYS.Base import SingletonArgs


@SingletonArgs
class XLRedisUtil(object):

    def __init__(self, department: BaseDPart, host, port, password):
        self.department = department
        # decode_responses=True 这样写存的数据是字符串格式
        pool = redis.ConnectionPool(host=host, port=port, decode_responses=True,  db=1,
                                    max_connections=4, password=password)
        self.r = redis.StrictRedis(connection_pool=pool)

        mq_pool = redis.ConnectionPool(host=host, port=port, decode_responses=True,  db=2,
                                    max_connections=15, password=password)
        self.mq_client = redis.StrictRedis(connection_pool=mq_pool)

        data_pool = redis.ConnectionPool(host=host, port=port, decode_responses=True,  db=3,
                                    max_connections=4, password=password)
        self.data_client = redis.StrictRedis(connection_pool=data_pool)

        pass

    def get_data_client(self):
        return self.data_client

    def get_mq_client(self):
        return self.mq_client

    def buildKey(self, key):
        return "{}:{}".format(self.department.name, key)

    def delete(self, key):
        self.r.delete(self.buildKey(key))

    def set(self, key, info, ex=None):
        self.r.set(self.buildKey(key), info, ex=ex)

    def acquire_lock(self, key, ex):
        if self.r.setnx(self.buildKey(key), key):
            self.r.expire(self.buildKey(key), ex)
            return True
        return False

    def get(self, key):
        return self.r.get(self.buildKey(key))

    def ttl(self, key):
        return self.r.ttl(self.buildKey(key))

    def incr(self, key, ex=None):
        t_key = self.buildKey(key)
        res =  self.r.incr(t_key)
        if ex is not None:
            self.r.expire(t_key, ex)
        return res

    def setDict(self, key, info: dict, ex=60):
        info_str = json.dumps(info, ensure_ascii=False)
        self.r.set(self.buildKey(key), info_str, ex=ex)

    def getDict(self, key) -> dict:
        info_str = self.r.get(self.buildKey(key))
        if info_str is None:
            return {}
        return json.loads(info_str)

    def listPut(self, key, info):
        self.r.lpush(self.buildKey(key), info)

    def listGetAll(self, key) -> dict:
        m_key = self.buildKey(key)
        t_size = self.r.llen(m_key)
        if t_size == 0:
            return []
        return self.r.lrange(m_key, 0, t_size)

    def hset(self, key,  vkey, vinfo):
        self.r.hset(self.buildKey(key), vkey, vinfo)

    def hget(self, key, vkey):
        return self.r.hget(self.buildKey(key), vkey)

    def hgetAll(self, key) -> dict:
        return self.r.hgetall(self.buildKey(key))

    def zadd(self, key,  info: dict):
        self.r.zadd(self.buildKey(key), info)

    def zscore(self, key,  value):
        return self.r.zscore(self.buildKey(key), value)

    def zcount(self, key, z_min=float("-inf"), z_max=float("inf")):
        return self.r.zcount(self.buildKey(key), z_min, z_max)

    def zrangebyscore(self, key, z_min=float("-inf"), z_max=float("inf")):
        return self.r.zrangebyscore(self.buildKey(key), z_min, z_max)


if __name__ == '__main__':
    red = XLRedisUtil()
    # red.set('sss', 'bbb')
    # print(red.get('sss'))
    # print(red.get('sssee'))
    # print(red.getDict('sssss'))
    # red.setDict('dsfs', {'ss': 'dd'})
    # print(red.getDict('dsfs'))
    red.zadd('ss', {'dd':12.3})
    print(red.zscore('ss','dd'))
    print(red.zcount('ss',z_min=20))
    print(red.zrangebyscore('ss'))



    pass



# coding=utf8
import re
import hashlib
import json
import torch
import pickle
import numpy
import struct


def to_readable_buffer(data):
    if isinstance(data, (bytes, bytearray, memoryview)):
        return data
    elif isinstance(data, str):
        return data.encode('utf-8')
    elif isinstance(data, int):
        return data.to_bytes((data.bit_length() + 7) // 8, byteorder='big')
    elif isinstance(data, float):
        return struct.pack('!f', data)
    elif isinstance(data, list):
        return b''.join(to_readable_buffer(item) for item in data)
    elif isinstance(data, dict):
        return json.dumps(data).encode('utf-8')
    elif isinstance(data, numpy.ndarray):
        return data.tobytes()
    elif isinstance(data, torch.Tensor):
        return to_readable_buffer(data.detach().cpu().numpy())
    else:
        return pickle.dumps(data)


def create_md5_hash(data):
    input_sha256 = hashlib.sha256()
    input_sha256.update(to_readable_buffer(data))
    data_hash = input_sha256.hexdigest()
    return data_hash


_find_unsafe = re.compile(r'[^\w@%+=:,./-]', re.ASCII).search


def quote(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return "''"
    if _find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return "'" + s.replace("'", "'\"'\"'") + "'"


def request_to_curl(request, compressed=False, verify=True):
    """
    Returns string with curl command by provided request object

    Parameters
    ----------
    compressed : bool
        If `True` then `--compressed` argument will be added to result
        :param verify:
        :param compressed:
        :param request:
    """
    parts = [
        ('curl', None),
        ('-X', request.method),
    ]

    for k, v in sorted(request.headers.items()):
        parts += [('-H', '{0}: {1}'.format(k, v))]

    if request.body:
        body = request.body
        if isinstance(body, bytes):
            body = body.decode('utf-8')
        parts += [('-d', body)]

    if compressed:
        parts += [('--compressed', None)]

    if not verify:
        parts += [('--insecure', None)]

    parts += [(None, request.url)]

    flat_parts = []
    for k, v in parts:
        if k:
            flat_parts.append(quote(k))
        if v:
            flat_parts.append(quote(v))

    return ' '.join(flat_parts)



# coding=utf8
import json
from typing import Dict

import requests

from ..Entities import XLDE, XLContext, XLBaseException, XLExceptionEnum
from ...Util import XLLogger
from ...Util.XLTools import request_to_curl


def doXLFuncPost(url, func_key, data_dict: Dict[str, XLDE], context: XLContext = None, env_dict: dict = None,
                 headers=None,
                 prn=True):
    logger = XLLogger()
    input_data_dict = {k: d.toLinkDE() for k, d in data_dict.items()}
    data = {'func_key': func_key,
            'data_dict': input_data_dict,
            'env_dict': env_dict
            }
    mheaders = {'http_debug': '1'}
    if headers is not None:
        mheaders.update(headers)
    response = requests.post(url, json=data, headers=mheaders)
    # 打印请求日志
    if context is not None:
        extra = {"mod": "http", "context_id": context.context_id,
                 "global_info": context.global_info.to_dict()}
    else:
        extra = {"mod": "http", "context_id": "", "global_info": ""}
    logger.requestLogger.info("{}".format(request_to_curl(response.request)), extra=extra)
    if prn:
        logger.infoContext(context, '{} {} {} {}'.format('doXLFuncPost', response.status_code, response.text,
                                                         json.dumps(data, ensure_ascii=False)))
    if response.status_code == 201:
        raise XLBaseException(XLExceptionEnum.http_debug_timeout)
    if response.status_code == 200:
        backInfo = json.loads(response.text)
        return backInfo.get('code'), backInfo.get('data_dict'), backInfo.get('env_dict')
    return -1, {}, {}
    pass


def doXLEventPost(url, event, data_dict: Dict[str, XLDE], want_keys=None, context: XLContext = None,
                  env_dict: dict = None, headers=None, prn=True):
    logger = XLLogger()
    input_data_dict = {k: d.toLinkDE() for k, d in data_dict.items()}
    want_keys = [] if want_keys is None else want_keys
    data = {'event': event,
            'data_dict': input_data_dict,
            'want_keys': want_keys,
            'env_dict': env_dict}
    mheaders = {'http_debug': '1'}
    if headers is not None:
        mheaders.update(headers)
    response = requests.post(url, json=data, headers=mheaders)
    # 打印请求日志
    if context is not None:
        extra = {"mod": "http", "context_id": context.context_id,
                 "global_info": context.global_info.to_dict()}
    else:
        extra = {"mod": "http", "context_id": "", "global_info": ""}
    logger.requestLogger.info("{}".format(request_to_curl(response.request)), extra=extra)
    if prn:
        logger.infoContext(context, '{} {} {} {}'.format('doXLEventPost', response.status_code, response.text,
                                                         json.dumps(data, ensure_ascii=False)))
    if response.status_code == 201:
        raise XLBaseException(XLExceptionEnum.http_debug_timeout)
    if response.status_code == 200:
        backInfo = json.loads(response.text)
        return backInfo.get('code'), backInfo.get('data_dict'), backInfo.get('env_dict')
    return -1, {}, {}
    pass


def doXLLearnPost(url, context_id, event, data_dict: Dict[str, XLDE], context: XLContext = None,
                  env_dict: dict = None, headers=None, prn=True):
    logger = XLLogger()
    input_data_dict = {k: d.toLinkDE() for k, d in data_dict.items()}
    data = {'event': event,
            'context_id': context_id,
            'data_dict': input_data_dict,
            'env_dict': env_dict}
    mheaders = {'http_debug': '1'}
    if headers is not None:
        mheaders.update(headers)
    response = requests.post(url, json=data, headers=mheaders)
    # 打印请求日志
    if context is not None:
        extra = {"mod": "http", "context_id": context.context_id,
                 "global_info": context.global_info.to_dict()}
    else:
        extra = {"mod": "http", "context_id": "", "global_info": ""}
    logger.requestLogger.info("{}".format(request_to_curl(response.request)), extra=extra)
    if prn:
        logger.infoContext(context, '{} {} {} {}'.format('doXLLearnPost', response.status_code, response.text,
                                                         json.dumps(data, ensure_ascii=False)))
    if response.status_code == 201:
        raise XLBaseException(XLExceptionEnum.http_debug_timeout)
    if response.status_code == 200:
        backInfo = json.loads(response.text)
        return backInfo.get('code'), backInfo.get('data_dict'), backInfo.get('env_dict')
    return -1, {}, {}
    pass


def doXLSelfPost(url, data_dict: dict, context: XLContext = None, headers=None, prn=True):
    logger = XLLogger()
    mheaders = {'http_debug': '1'}
    if headers is not None:
        mheaders.update(headers)
    response = requests.post(url, json=data_dict, headers=mheaders)
    # 打印请求日志
    if context is not None:
        extra = {"mod": "http", "context_id": context.context_id,
                 "global_info": context.global_info.to_dict()}
    else:
        extra = {"mod": "http", "context_id": "", "global_info": ""}
    logger.requestLogger.info("{}".format(request_to_curl(response.request)), extra=extra)
    if prn:
        logger.infoContext(context, '{} {} {} {}'.format('doXLSelfPost', response.status_code, response.text,
                                                         json.dumps(data_dict, ensure_ascii=False)))
    if response.status_code == 201:
        raise XLBaseException(XLExceptionEnum.http_debug_timeout)
    if response.status_code == 200:
        backInfo = json.loads(response.text)
        return backInfo.get('code'), backInfo.get('data_dict'), backInfo.get('env_dict')
    return -1, {}, {}
    pass




# coding=utf8
import json
import time
import traceback
from concurrent.futures import ThreadPoolExecutor

import requests
from redis import Redis

from ...Department import BaseDPart
from ...Util import XLRedisUtil, XLLogger

dev_mq_key = None
online_mq_key = None


def _getSetSubIndex(redis_client: Redis, name, group, new_id=None):
    key = group + ":" + name + "readid"
    if new_id is not None:
        redis_client.set(key, new_id)
        return new_id
    return redis_client.get(key) or '0-0'


def _mqHandler(rec_mq_key, department: BaseDPart, env_flag):
    logger = XLLogger()
    redis_util = XLRedisUtil()
    redis_client = redis_util.get_mq_client()
    try:
        group = 'main'
        try:
            redis_client.xinfo_consumers(name=rec_mq_key, groupname=group)
        except Exception as e:
            redis_client.xadd(rec_mq_key, {'sys': 'create'})
            consumer_id = _getSetSubIndex(redis_client, group=group, name=rec_mq_key)
            redis_client.xgroup_create(name=rec_mq_key, groupname=group, id=consumer_id)

        while True:
            content = ""
            try:
                # block 0 时阻塞等待, 其他数值表示读取超时时间
                [[name, [(nid, content)]]] = redis_client.xreadgroup(groupname=group, consumername="consumer",
                                                                     streams={rec_mq_key: '>'}, count=1, block=0)
                _ = _getSetSubIndex(redis_client, group=group, name=rec_mq_key, new_id=nid)
                if not isinstance(content, dict) or len(content) == 0 or content.get('sys') is not None:
                    continue
                if env_flag == 'dev':
                    # if rec_mq_key == 'dev_http_event':
                    b_url = 'http://127.0.0.1:{}{}'.format(department.port, content['path'])
                    # 测试环境执行的逻辑
                    if content['method'] == 'POST':
                        response = requests.post(b_url, json=json.loads(content['str_body']))
                        redis_client.xadd(online_mq_key,
                                          {'evnet_id': content['evnet_id'], 'content': response.content,
                                           'status_code': response.status_code, 'text': response.text})
                    elif content['method'] == 'GET':
                        response = requests.get(b_url, params=json.loads(content['path_params']))
                        redis_client.xadd(online_mq_key,
                                          {'evnet_id': content['evnet_id'],
                                           'status_code': response.status_code,
                                           'content': response.content,
                                           'text': response.text})
                else:
                    # 运行环境执行的逻辑; 使用默认的保存60s
                    redis_util.setDict('dev_http_back_' + content['evnet_id'], content)
                pass
            except Exception as ex:
                logger.error("mq消息处理异常 content:{} e:{}".format(content, ex))
                traceback.print_exc()
                pass
    except Exception as e:
        logger.error("mq _mqHandlerDev 启动异常 e:{}".format(e))
        traceback.print_exc()
    pass


def _mqdev_handler(department: BaseDPart, env_flag):
    logger = XLLogger()
    if env_flag != 'dev':
        logger.info('_mqdev_handler 环境不是dev 可以不用执行')
        return
    while True:
        try:
            response = requests.post('http://host.home.tokgo.cn:{}/dev/mq/http/link/register/'.format(department.port))
            # 间隔50秒
            time.sleep(50)
            pass
        except Exception as e:
            logger.error('_mqdev_handler 处理任务错')
            traceback.print_exc()
            time.sleep(30)
            pass
    pass


def sendDevMqData(evnet_id, method, path, path_params, str_body):
    try:
        return XLRedisUtil().get_mq_client().xadd(dev_mq_key, {'evnet_id': evnet_id, 'method': method, 'path': path,
                                                               'path_params': json.dumps(path_params, ensure_ascii=False),
                                                               'str_body': str_body})
    except Exception as e:
        logger = XLLogger()
        logger.error("mq http 发送消息失败; e:{}".format(e))
    pass


def HttpDevMQInit(department: BaseDPart, env_flag='dev'):
    global dev_mq_key
    global online_mq_key
    dev_mq_key = '{}_dev_http_event'.format(department.name)
    online_mq_key = '{}_online_http_event'.format(department.name)

    threadPool = ThreadPoolExecutor(max_workers=2)
    if env_flag == 'dev':
        threadPool.submit(_mqHandler, dev_mq_key, department, env_flag)
        # threadPool.submit(_mqHandler, 'online_http_event')
    else:
        threadPool.submit(_mqHandler, online_mq_key, department, env_flag)
    # dev 环境相关的处理线程
    threadPool.submit(_mqdev_handler, department, env_flag)
    pass



# coding=utf8
from pydantic import BaseModel

from ..Base import AISerializeObject


class XLLinkDM(BaseModel):
    key: str
    data_type: str


class XLDM(AISerializeObject):
    def __init__(self, key, data_type, des=None):
        # 类型
        self.data_type = data_type
        # key
        self.key = key
        # 描述
        self.des = key if des is None else des

    def __str__(self):
        return '<{}><{}>'.format(self.key, self.data_type)
        pass

    def toLinkDM(self):
        return {
            'key': self.key,
            'data_type': self.data_type
        }
        pass


class XLDataMarkUtil(object):

    @staticmethod
    def keyDataBuild(key, data_type) -> XLDM:
        entity: XLDM = XLDM(key=key, data_type=data_type)
        entity.association_chain = [0, []]
        entity.create_identity = 0
        return entity
        pass


    @staticmethod
    def list2DictByKey(data_list: list) -> dict:
        data_dict = {}
        for d in data_list:
            data_dict[d.key] = d
        return data_dict
        pass


pass



# coding=utf8
import threading

'''
支持多线程的单例
'''


class SingletonArgs(object):

    def __init__(self, cls):
        self._instance_lock = threading.Lock()
        self._cls = cls
        self.uniqueInstance = None

    def __call__(self, **kwargs):
        if self.uniqueInstance is None:
            with self._instance_lock:
                if self.uniqueInstance is None:
                    self.uniqueInstance = self._cls(**kwargs)
        return self.uniqueInstance



# coding=utf8
import copy
from datetime import datetime

from ..Base import AIFuncSerialize
from ...Department import BaseDPart


class XLCGlobalInfo(AIFuncSerialize):
    def __init__(self, context_id, up_globalInfo: dict):
        if up_globalInfo is None or up_globalInfo.get('all_trace_id') is None:
            self.all_trace_id = []
        else:
            self.all_trace_id = copy.deepcopy(up_globalInfo['all_trace_id'])

        self.all_trace_id.append(context_id)
        pass

    def to_dict(self):
        return {
            'all_trace_id': self.all_trace_id
        }


class XLContext(AIFuncSerialize):
    def __init__(self, department: BaseDPart, part, up_globalInfo: dict):
        department_name = department.name.lower()
        self.context_id = datetime.now().strftime("{}-{}-%Y%m%d_%H%M%S_%f".format(department_name, part))
        self.global_info = XLCGlobalInfo(context_id=self.context_id, up_globalInfo=up_globalInfo)
        self.run_data_pool = {}
        self.run_trace_list = []
        pass

    pass


pass
